library ieee;
use ieee.numeric_std.all;
use ieee.std_logic_1164.all;
use work.quarkpkg.all;

entity quark is

  port (
    ClkxCI   : in  std_logic;
    RstxRBI  : in  std_logic;
    StartxSI : in  std_logic;
    INENxEI  : in  std_logic;
    OUTENxEO : out std_logic;
    DxDI     : in  std_logic_vector(RATE*8-1 downto 0);
    DxDO     : out std_logic_vector(RATE*8-1 downto 0));

end quark;

architecture rtl of quark is

  component controller
    port (
      ClkxCI    : in  std_logic;
      RstxRBI   : in  std_logic;
      StartxSI  : in  std_logic;
      INENxEI   : in  std_logic;
      FreezexSO : out std_logic;
      OUTENxEO  : out std_logic);
  end component;

  signal QuarkStatexDP, QuarkStatexDN : std_logic_vector(0 to WWIDTH*8-1);
  signal PermOutxD                    : std_logic_vector(0 to WWIDTH*8-1);
  signal X, Y                         : std_logic_vector(0 to WWIDTH*8/2-1);
  signal LxDN, LxDP, LoutxD           : std_logic_vector(0 to LWIDTH-1);
  signal Xn, Yn, Ln, h, Xnn, Ynn      : std_logic;
  signal OUTENxE, FreezexS            : std_logic;
  signal DInxD                        : std_logic_vector(0 to RATE*8-1);
  signal IR, IW, ID                   : std_logic_vector(7 downto 0);
  
begin  -- rtl

  f_inout : for i in 0 to RATE*8-1 generate
    DInxD(i)         <= DxDI(i);
    DxDO(RATE*8-1-i) <= PermOutxD(8*(WWIDTH-RATE)+i);
  end generate f_inout;

  OUTENxEO <= OUTENxE;

  -- FSM
  -----------------------------------------------------------------------------
  u_controller : controller
    port map (
      ClkxCI    => ClkxCI,
      RstxRBI   => RstxRBI,
      StartxSI  => StartxSI,
      INENxEI   => INENxEI,
      FreezexSO => FreezexS,
      OUTENxEO  => OUTENxE);

  -- Input Select
  -----------------------------------------------------------------------------
  QuarkStatexDN(0 to 8*(WWIDTH-RATE)-1) <= INITSTATE(0 to 8*(WWIDTH-RATE)-1) when StartxSI = '1' and INENxEI = '1' else
                                           PermOutxD(0 to 8*(WWIDTH-RATE)-1);

  QuarkStatexDN(8*(WWIDTH-RATE) to 8*WWIDTH-1) <= INITSTATE(8*(WWIDTH-RATE) to 8*WWIDTH-1) xor DinxD when StartxSI = '1' and INENxEI = '1' else
                                                  PermOutxD(8*(WWIDTH-RATE) to 8*WWIDTH-1) xor DinxD when StartxSI = '0' and INENxEI = '1' else
                                                  PermOutxD(8*(WWIDTH-RATE) to 8*WWIDTH-1);

  LxDN <= (others => '1') when INENxEI = '1' or StartxSI = '1' or OUTENxE = '1' else LoutxD;

  -- Permutation
  -----------------------------------------------------------------------------
  X <= QuarkStatexDP(0 to WWIDTH*8/2-1);
  Y <= QuarkStatexDP(WWIDTH*8/2 to WWIDTH*8-1);

  Xn <= Y(0) xor X(0) xor X(9) xor X(14) xor X(21) xor X(28) xor
        X(33) xor X(37) xor X(45) xor X(52) xor X(55) xor X(50) xor
        (X(59) and X(55)) xor (X(37) and X(33)) xor (X(15) and X(9)) xor
        (X(55) and X(52) and X(45)) xor (X(33) and X(28) and X(21)) xor
        (X(59) and X(45) and X(28) and X(9)) xor
        (X(55) and X(52) and X(37) and X(33)) xor
        (X(59) and X(55) and X(21) and X(15)) xor
        (X(59) and X(55) and X(52) and X(45) and X(37)) xor
        (X(33) and X(28) and X(21) and X(15) and X(9)) xor
        (X(52) and X(45) and X(37) and X(33) and X(28) and X(21));

  Yn <= Y(0) xor Y(7) xor Y(16) xor Y(20) xor Y(30) xor
        Y(35) xor Y(37) xor Y(42) xor Y(51) xor Y(54) xor Y(49) xor
        (Y(58) and Y(54)) xor (Y(37) and Y(35)) xor (Y(15) and Y(7)) xor
        (Y(54) and Y(51) and Y(42)) xor (Y(35) and Y(30) and Y(20)) xor
        (Y(58) and Y(42) and Y(30) and Y(7)) xor
        (Y(54) and Y(51) and Y(37) and Y(35)) xor
        (Y(58) and Y(54) and Y(20) and Y(15)) xor
        (Y(58) and Y(54) and Y(51) and Y(42) and Y(37)) xor
        (Y(35) and Y(30) and Y(20) and Y(15) and Y(7)) xor
        (Y(51) and Y(42) and Y(37) and Y(35) and Y(30) and Y(20));

  Ln <= LxDP(0) xor LxDP(3);

  h <= LxDP(0) xor X(1) xor Y(2) xor X(4) xor Y(10) xor X(25) xor X(31) xor Y(43) xor X(56) xor Y(59) xor
       (Y(3) and X(55)) xor (X(46) and X(55)) xor (X(55) and Y(59)) xor (Y(3) and X(25) and X(46)) xor (Y(3) and X(46) and X(55)) xor
       (Y(3) and X(46) and Y(59)) xor (LxDP(0) and X(25) and X(46) and Y(59)) xor (LxDP(0) and X(25));
  
  Xnn <= Xn xor h;
  Ynn <= Yn xor h;

  PermOutxD <= QuarkStatexDP(1 to WWIDTH*8/2-1) & Xnn & QuarkStatexDP(WWIDTH*8/2+1 to WWIDTH*8-1) & Ynn;
  LoutxD    <= LxDP(1 to LWIDTH-1) & Ln;

  -- State Register
  -----------------------------------------------------------------------------
  p_mem : process (ClkxCI, RstxRBI)
  begin  -- process p_mem
    if RstxRBI = '0' then               -- asynchronous reset (active low)
      QuarkStatexDP <= (others => '0');
      LxDP          <= (others => '0');
      
    elsif ClkxCI'event and ClkxCI = '1' then  -- rising clock edge
      --if FreezexS = '0' then
      QuarkStatexDP <= QuarkStatexDN;
      LxDP          <= LxDN;

      --end if;
      
    end if;
  end process p_mem;

end rtl;
