library ieee;
use ieee.numeric_std.all;
use ieee.std_logic_1164.all;
use work.quarkpkg.all;

entity quark is

  port (
    ClkxCI   : in  std_logic;
    RstxRBI  : in  std_logic;
    StartxSI : in  std_logic;
    INENxEI  : in  std_logic;
    OUTENxEO : out std_logic;
    DxDI     : in  std_logic_vector(RATE*8-1 downto 0);
    DxDO     : out std_logic_vector(RATE*8-1 downto 0));

end quark;

architecture rtl of quark is

  component controller
    port (
      ClkxCI    : in  std_logic;
      RstxRBI   : in  std_logic;
      StartxSI  : in  std_logic;
      INENxEI   : in  std_logic;
      FreezexSO : out std_logic;
      OUTENxEO  : out std_logic);
  end component;

  signal QuarkStatexDP, QuarkStatexDN : std_logic_vector(0 to WWIDTH*8-1);
  signal PermOutxD                    : std_logic_vector(0 to WWIDTH*8-1);
  signal X, Y                         : std_logic_vector(0 to WWIDTH*8/2-1);
  signal LxDN, LxDP, LoutxD           : std_logic_vector(0 to LWIDTH-1);
  signal Xn, Yn, Ln, h, Xnn, Ynn      : std_logic_vector(0 to PDEG-1);
  signal OUTENxE, FreezexS            : std_logic;
  signal DInxD                        : std_logic_vector(0 to RATE*8-1);
  signal IR, IW, ID                   : std_logic_vector(7 downto 0);
  
begin  -- rtl

  f_inout : for i in 0 to RATE*8-1 generate
    DInxD(i)         <= DxDI(i);
    DxDO(RATE*8-1-i) <= PermOutxD(8*(WWIDTH-RATE)+i);
  end generate f_inout;

  OUTENxEO <= OUTENxE;

  -- FSM
  -----------------------------------------------------------------------------
  u_controller : controller
    port map (
      ClkxCI    => ClkxCI,
      RstxRBI   => RstxRBI,
      StartxSI  => StartxSI,
      INENxEI   => INENxEI,
      FreezexSO => FreezexS,
      OUTENxEO  => OUTENxE);

  -- Input Select
  -----------------------------------------------------------------------------
  QuarkStatexDN(0 to 8*(WWIDTH-RATE)-1) <= INITSTATE(0 to 8*(WWIDTH-RATE)-1) when StartxSI = '1' and INENxEI = '1' else
                                           PermOutxD(0 to 8*(WWIDTH-RATE)-1);

  QuarkStatexDN(8*(WWIDTH-RATE) to 8*WWIDTH-1) <= INITSTATE(8*(WWIDTH-RATE) to 8*WWIDTH-1) xor DinxD when StartxSI = '1' and INENxEI = '1' else
                                                  PermOutxD(8*(WWIDTH-RATE) to 8*WWIDTH-1) xor DinxD when StartxSI = '0' and INENxEI = '1' else
                                                  PermOutxD(8*(WWIDTH-RATE) to 8*WWIDTH-1);
  
  LxDN <= (others => '1') when INENxEI = '1' or StartxSI = '1' or OUTENxE = '1' else LoutxD;

  -- Permutation
  -----------------------------------------------------------------------------

  X <= QuarkStatexDP(0 to WWIDTH*8/2-1);
  Y <= QuarkStatexDP(WWIDTH*8/2 to WWIDTH*8-1);

  perm : for i in 0 to PDEG-1 generate

    Xn(i) <= Y(0+i) xor X(0+i) xor X(9+i) xor X(14+i) xor X(21+i) xor X(28+i) xor
             X(33+i) xor X(37+i) xor X(45+i) xor X(52+i) xor X(55+i) xor X(50+i) xor
             (X(59+i) and X(55+i)) xor (X(37+i) and X(33+i)) xor (X(15+i) and X(9+i)) xor
             (X(55+i) and X(52+i) and X(45+i)) xor (X(33+i) and X(28+i) and X(21+i)) xor
             (X(59+i) and X(45+i) and X(28+i) and X(9+i)) xor
             (X(55+i) and X(52+i) and X(37+i) and X(33+i)) xor
             (X(59+i) and X(55+i) and X(21+i) and X(15+i)) xor
             (X(59+i) and X(55+i) and X(52+i) and X(45+i) and X(37+i)) xor
             (X(33+i) and X(28+i) and X(21+i) and X(15+i) and X(9+i)) xor
             (X(52+i) and X(45+i) and X(37+i) and X(33+i) and X(28+i) and X(21+i));
    
    Yn(i) <= Y(0+i) xor Y(7+i) xor Y(16+i) xor Y(20+i) xor Y(30+i) xor
             Y(35+i) xor Y(37+i) xor Y(42+i) xor Y(51+i) xor Y(54+i) xor Y(49+i) xor
             (Y(58+i) and Y(54+i)) xor (Y(37+i) and Y(35+i)) xor (Y(15+i) and Y(7+i)) xor
             (Y(54+i) and Y(51+i) and Y(42+i)) xor (Y(35+i) and Y(30+i) and Y(20+i)) xor
             (Y(58+i) and Y(42+i) and Y(30+i) and Y(7+i)) xor
             (Y(54+i) and Y(51+i) and Y(37+i) and Y(35+i)) xor
             (Y(58+i) and Y(54+i) and Y(20+i) and Y(15+i)) xor
             (Y(58+i) and Y(54+i) and Y(51+i) and Y(42+i) and Y(37+i)) xor
             (Y(35+i) and Y(30+i) and Y(20+i) and Y(15+i) and Y(7+i)) xor
             (Y(51+i) and Y(42+i) and Y(37+i) and Y(35+i) and Y(30+i) and Y(20+i));

    h(i) <= LxDP(i) xor X(1+i) xor Y(2+i) xor X(4+i) xor Y(10+i) xor X(25+i) xor X(31+i) xor Y(43+i) xor X(56+i) xor Y(59+i) xor
            (Y(3+i) and X(55+i)) xor (X(46+i) and X(55+i)) xor (X(55+i) and Y(59+i)) xor (Y(3+i) and X(25+i) and X(46+i)) xor
            (Y(3+i) and X(46+i) and X(55+i)) xor
            (Y(3+i) and X(46+i) and Y(59+i)) xor (LxDP(i) and X(25+i) and X(46+i) and Y(59+i)) xor (LxDP(i) and X(25+i));
    
    Xnn(i) <= Xn(i) xor h(i);
    Ynn(i) <= Yn(i) xor h(i);
    
  end generate perm;

  Ln(0) <= LxDP(0) xor LxDP(3);
  Ln(1) <= LxDP(1) xor LxDP(4);
  Ln(2) <= LxDP(2) xor LxDP(5);
  Ln(3) <= LxDP(3) xor LxDP(6);
  Ln(4) <= LxDP(4) xor LxDP(7);
  Ln(5) <= LxDP(5) xor LxDP(8);
  Ln(6) <= LxDP(6) xor LxDP(9);
  Ln(7) <= LxDP(7) xor Ln(0);


  PermOutxD <= QuarkStatexDP(PDEG to WWIDTH*8/2-1) & Xnn & QuarkStatexDP(WWIDTH*8/2+PDEG to WWIDTH*8-1) & Ynn;
  LoutxD    <= LxDP(PDEG to LWIDTH-1) & Ln;

  -- State Register
  -----------------------------------------------------------------------------
  p_mem : process (ClkxCI, RstxRBI)
  begin  -- process p_mem
    if RstxRBI = '0' then               -- asynchronous reset (active low)
      QuarkStatexDP <= (others => '0');
      LxDP          <= (others => '0');
      
    elsif ClkxCI'event and ClkxCI = '1' then  -- rising clock edge
      --if FreezexS = '0' then
      QuarkStatexDP <= QuarkStatexDN;
      LxDP          <= LxDN;

      --end if;
      
    end if;
  end process p_mem;

end rtl;
